// TrackTrek '97 by Steven Lawrance
// Database Object (Server Version)

package sl.TrackTrek;

import java.io.*;
import java.util.*;

public abstract class DBObject extends Object
{
  public final static int OK = 1;
  public final static int ERROR = 2;
  public final static int PING = 3;
  public final static int READ = 4;
  public final static int WRITE = 5;
  public final static int CLOSE = 6;
  public final static int LOCK = 7;
  public final static int NOLOCK = 9;
  public final static int NOTEXIST = 10;
  public final static int UNLOCK = 11;

  // unimplemented (user access messages; right now, everyone has full access)
  public final static int DENIED = 20;
  public final static int AUTH = 25;

  public final static int DATE = 1;
  public final static int DIST = 2;
  public final static int STR = 3;
  public final static int TIME = 4;
  public final static int EVENT = 5;
  public final static int TEAM = 6;
  public final static int TRACKMEET = 7;
  public final static int PARTICIPANT = 8;
  public final static int USER = 9;
  public final static int GROUP = 10;
  public final static int ROOT = 11;
  public final static int FILE = 12;

  public Vector locks;
  public Object exclusive;
  public boolean lockChildren;
  public int version;

  public DBObject()
  {
    locks = new Vector();
    exclusive = null;
    lockChildren = false;
    version = 0;
  }

  // Versions
  public int getVersion()
  {
    return version;
  }

  public int newVersion()
  {
    if (version > 500000)
      version = 0;
    return ++version;
  }

  // Initialization/Storage
  public abstract void load(DataInputStream in, boolean includeChildren);
  public abstract void save(DataOutputStream out, boolean includeChildren);

  // Data type
  public abstract int dataType();

  // Pointer manipulators       
  public static String rewind(String current)
  {
    // Return the parent of the current folder
    int i = current.lastIndexOf('|');
    if (i >= 0)
      return current.substring(0, i);
    else
      return null;
  }

/* public static String forward(String current)
  {
    // Remove the first folder
    int loc = current.indexOf('|');
    if (loc >= 0)
      return current.substring(loc + 1);
    else
      return null;
  }

  public static String getFolderName(String current)
  {
    // Returns the name of the first folder
    int loc = current.indexOf('|');
    if (loc >= 0)
      return current.substring(0, loc);
    else
      return current;
  }*/

  public static DBObject objectFromPointer(DBObject root, String location)
  {
    // Go through and get the object
    StringTokenizer next = new StringTokenizer(location, "|", false);
    DBObject obj = root;
    while (next.hasMoreTokens())
    {
      // Get the next child
      obj = obj.getChild(next);

      // If the object was not found, return null
      if (obj == null)
        return null;
    }

    // Return the object
    return obj;
  }

  public abstract DBObject getChild(StringTokenizer location);


  // Locking mechanism-----------------------

  public void lockInfo(DataOutputStream out)
  {
    try
    {
      // Return information on our current locking status
      out.writeInt(locks.size());
      out.writeBoolean(((exclusive != null)? true: false));
      out.writeBoolean(lockChildren);
    } catch (IOException str) {}
  }

  public void forceUnlock()
  {
    // Forces a complete unlock of this object
    locks.removeAllElements();
    exclusive = null;
    lockChildren = false;
  }

  public abstract int lockDBChildren(Object lock, boolean exclusive, boolean waitFlag, int waitDuration);
  public abstract int unlockDBChildren(Object lock);

  public static int checkParents(DBObject root, String path)
  {
    String current = rewind(path);
    int access = LOCK;
    while ((access == LOCK) && (current != null))
    {
      // Check parent's lockChildren flag; if set, we can't have access :-(
      DBObject obj = objectFromPointer(root, current);
      if (obj != null)
        if (obj.lockChildren)
          access = NOLOCK;
      current = rewind(path);
    }

    // Check root and return the final result
    if (access == LOCK)
      access = ((root.lockChildren)? NOLOCK: LOCK);
    return access;
  }

  public int unlockObject(Object lock, boolean unlockChildren)
  {
    // Determine the type of lock this has, if any
    if (lock.equals(exclusive))
    {
      // This object has exclusive access; free the children first
      if (lockChildren)
        unlockDBChildren(lock);

      // Free this object
      exclusive = null;
      lockChildren = false;

      // Return the new status
      return NOLOCK;
    }

    if (locks.contains(lock))
    {
      // This lock is a read lock; if we need to unlock the children, do so
      if (unlockChildren)
        unlockDBChildren(lock);

      // Remove our local lock
      locks.removeElement(lock);

      // Return our current state
      return NOLOCK;
    }

    // Return its non-existence
    return NOTEXIST;
  }

  public int lockObject(Object lock, boolean exclusive, boolean lockChildren, boolean waitFlag, int waitDuration)
  {
    // This function assumes that the parents had been checked for access
    if (exclusive)
    {
      // The client wants exclusive access; check to see if it is already set
      if (this.exclusive == null)
      {
        // Wait until there are no read locks
        long startMs = System.currentTimeMillis();
        boolean freed = locks.isEmpty();
        while ((startMs + waitDuration > System.currentTimeMillis()) && !freed)
        {
          freed = locks.isEmpty();
          Thread.yield();
        }

        if (freed)
        {
          // We were able to get exclusive access :-)
          this.exclusive = lock;
          this.lockChildren = lockChildren;
        }
        else
        {
          // Could not obtain exclusive access
          return NOLOCK;
        }
      }
      else
      {
        if (waitFlag)
        {
          // Object is locked; wait for a while until we have access
          long startMs = System.currentTimeMillis();
          boolean freed = false;
          while ((startMs + waitDuration > System.currentTimeMillis()) && !freed)
          {
            if (this.exclusive == null)
              freed = true;
            Thread.yield();
          }

          // If the object was eventually freed, then check the read locks
          if (freed)
          {
            // Wait until there are no read locks
            startMs = System.currentTimeMillis();
            freed = locks.isEmpty();
            while ((startMs + waitDuration > System.currentTimeMillis()) && !freed)
            {
              freed = locks.isEmpty();
              Thread.yield();
            }

            if (freed)
            {
              // We were able to get exclusive access :-)
              this.exclusive = lock;
              this.lockChildren = lockChildren;
            }
            else
            {
              // Could not obtain exclusive access
              return NOLOCK;
            }
          }
          else
          {
            // Object is still locked :-(
            return NOLOCK;
          }
        }
        else
        {
          // Object is locked, but we won't wait for it to free
          return NOLOCK;
        }
      }
    }

    // Lock desired is "non-exclusive" access
    else
    {
      if (this.exclusive != null)
      {
        // Object is currently in exclusive mode
        if (waitFlag)
        {
          // Wait for a while until we have access
          long startMs = System.currentTimeMillis();
          boolean freed = false;
          while ((startMs + waitDuration > System.currentTimeMillis()) && !freed)
          {
            if (this.exclusive == null)
              freed = true;
            Thread.yield();
          }

          // If the object was eventually freed, then add a read lock
          if (freed)
          {
            locks.addElement(lock);
          }
          else
          {
            // Object is still locked :-(
            return NOLOCK;
          }
        }
        else
        {
          // We didn't wait; notify the client that it was locked
          return NOLOCK;
        }
      }

      // No one has exclusive access; add this lock to the read locks
      else
      {
        locks.addElement(lock);
      }
    }

    // Lock the children, if any
    int success = lockDBChildren(lock, exclusive, waitFlag, waitDuration);

    // If the children could not be locked, then unlock this object
    if (success != LOCK)
    {
      if (exclusive)
      {
        this.exclusive = null;
        this.lockChildren = false;
      }
      else
        locks.removeElement(lock);

      // Unlock any of our children
      unlockDBChildren(lock);
      return NOLOCK;
    }

    // The desired lock was successful :-)!!
    return LOCK;
  }
}
