// TrackTrek '97 by Steven Lawrance
// Database Server

// This program "keeps track" of all the information

import java.net.*;
import java.io.*;
import java.util.*;
import java.awt.*;


// Database Object Classes----------------------------

abstract class DBObject extends Object
{
  public final static int OK = 1;
  public final static int ERROR = 2;
  public final static int PING = 3;
  public final static int READ = 4;
  public final static int WRITE = 5;
  public final static int CLOSE = 6;
  public final static int LOCK = 7;
  public final static int NOLOCK = 9;
  public final static int NOTEXIST = 10;
  public final static int UNLOCK = 11;
  public final static int EXIST = 12;
  public final static int ADD = 14;
  public final static int RENAME = 15;
  public final static int DELETE = 16;

  // unimplemented (user access messages; right now, everyone has full access)
  public final static int DENIED = 20;
  public final static int AUTH = 25;

  public final static int DATE = 1;
  public final static int DIST = 2;
  public final static int STR = 3;
  public final static int TIME = 4;
  public final static int EVENT = 5;
  public final static int TEAM = 6;
  public final static int TRACKMEET = 7;
  public final static int PARTICIPANT = 8;
  public final static int USER = 9;
  public final static int GROUP = 10;
  public final static int ROOT = 11;
  public final static int FILE = 12;

  public Vector locks;
  public Object exclusive;
  public boolean lockChildren;
  public int version;

  public DBObject()
  {
    locks = new Vector();
    exclusive = null;
    lockChildren = false;
    version = 0;
  }

  // Versions
  public int getVersion()
  {
    return version;
  }

  public int newVersion()
  {
    if (version > 500000)
      version = 0;
    return ++version;
  }

  // Initialization/Storage
  public abstract void load(DataInputStream in, boolean includeChildren);
  public abstract void save(DataOutputStream out, boolean includeChildren);

  // Data type
  public abstract int dataType();

  // Pointer manipulators       
  public static String rewind(String current)
  {
    // Return the parent of the current folder
    int i = current.lastIndexOf('|');
    if (i >= 0)
      return current.substring(0, i);
    else
      return null;
  }

/* public static String forward(String current)
  {
    // Remove the first folder
    int loc = current.indexOf('|');
    if (loc >= 0)
      return current.substring(loc + 1);
    else
      return null;
  }

  public static String getFolderName(String current)
  {
    // Returns the name of the first folder
    int loc = current.indexOf('|');
    if (loc >= 0)
      return current.substring(0, loc);
    else
      return current;
  }*/

  public static DBObject objectFromPointer(DBObject root, String location)
  {
    // Go through and get the object
    StringTokenizer next = new StringTokenizer(location, "|", false);
    DBObject obj = root;
    while (next.hasMoreTokens())
    {
      // Get the next child
      obj = obj.getChild(next);

      // If the object was not found, return null
      if (obj == null)
        return null;
    }

    // Return the object
    return obj;
  }

  public abstract DBObject getChild(StringTokenizer location);

  // Add/Rename/Delete-----------------------
  
  public abstract int add(ttStr name, int type, DataInputStream in, DataOutputStream out, boolean addChildren);
  public abstract int rename(ttStr oldName, ttStr newName);
  public abstract int delete(ttStr name);

  // Locking mechanism-----------------------

  public void lockInfo(DataOutputStream out)
  {
    try
    {
      // Return information on our current locking status
      out.writeInt(locks.size());
      out.writeBoolean(((exclusive != null)? true: false));
      out.writeBoolean(lockChildren);
    } catch (IOException str) {}
  }

  public void forceUnlock()
  {
    // Forces a complete unlock of this object
    locks.removeAllElements();
    exclusive = null;
    lockChildren = false;
  }

  public abstract int lockDBChildren(Object lock, boolean exclusive, boolean waitFlag, int waitDuration);
  public abstract int unlockDBChildren(Object lock);

  public static int checkParents(DBObject root, String path)
  {
    String current = rewind(path);
    int access = LOCK;
    while ((access == LOCK) && (current != null))
    {
      // Check parent's lockChildren flag; if set, we can't have access :-(
      DBObject obj = objectFromPointer(root, current);
      if (obj != null)
        if (obj.lockChildren)
          access = NOLOCK;
      current = rewind(path);
    }

    // Check root and return the final result
    if (access == LOCK)
      access = ((root.lockChildren)? NOLOCK: LOCK);
    return access;
  }

  public int unlockObject(Object lock, boolean unlockChildren)
  {
    // Determine the type of lock this has, if any
    if (lock.equals(exclusive))
    {
      // This object has exclusive access; free the children first
      if (lockChildren)
        unlockDBChildren(lock);

      // Free this object
      exclusive = null;
      lockChildren = false;

      // Return the new status
      return NOLOCK;
    }

    if (locks.contains(lock))
    {
      // This lock is a read lock; if we need to unlock the children, do so
      if (unlockChildren)
        unlockDBChildren(lock);

      // Remove our local lock
      locks.removeElement(lock);

      // Return our current state
      return NOLOCK;
    }

    // Return its non-existence
    return NOTEXIST;
  }

  public int lockObject(Object lock, boolean exclusive, boolean lockChildren, boolean waitFlag, int waitDuration)
  {
    // This function assumes that the parents had been checked for access
    if (exclusive)
    {
      // The client wants exclusive access; check to see if it is already set
      if (this.exclusive == null)
      {
        // Wait until there are no read locks
        long startMs = System.currentTimeMillis();
        boolean freed = locks.isEmpty();
        while ((startMs + waitDuration > System.currentTimeMillis()) && !freed)
        {
          freed = locks.isEmpty();
          Thread.yield();
        }

        if (freed)
        {
          // We were able to get exclusive access :-)
          this.exclusive = lock;
          this.lockChildren = lockChildren;
        }
        else
        {
          // Could not obtain exclusive access
          return NOLOCK;
        }
      }
      else
      {
        if (waitFlag)
        {
          // Object is locked; wait for a while until we have access
          long startMs = System.currentTimeMillis();
          boolean freed = false;
          while ((startMs + waitDuration > System.currentTimeMillis()) && !freed)
          {
            if (this.exclusive == null)
              freed = true;
            Thread.yield();
          }

          // If the object was eventually freed, then check the read locks
          if (freed)
          {
            // Wait until there are no read locks
            startMs = System.currentTimeMillis();
            freed = locks.isEmpty();
            while ((startMs + waitDuration > System.currentTimeMillis()) && !freed)
            {
              freed = locks.isEmpty();
              Thread.yield();
            }

            if (freed)
            {
              // We were able to get exclusive access :-)
              this.exclusive = lock;
              this.lockChildren = lockChildren;
            }
            else
            {
              // Could not obtain exclusive access
              return NOLOCK;
            }
          }
          else
          {
            // Object is still locked :-(
            return NOLOCK;
          }
        }
        else
        {
          // Object is locked, but we won't wait for it to free
          return NOLOCK;
        }
      }
    }

    // Lock desired is "non-exclusive" access
    else
    {
      if (this.exclusive != null)
      {
        // Object is currently in exclusive mode
        if (waitFlag)
        {
          // Wait for a while until we have access
          long startMs = System.currentTimeMillis();
          boolean freed = false;
          while ((startMs + waitDuration > System.currentTimeMillis()) && !freed)
          {
            if (this.exclusive == null)
              freed = true;
            Thread.yield();
          }

          // If the object was eventually freed, then add a read lock
          if (freed)
          {
            locks.addElement(lock);
          }
          else
          {
            // Object is still locked :-(
            return NOLOCK;
          }
        }
        else
        {
          // We didn't wait; notify the client that it was locked
          return NOLOCK;
        }
      }

      // No one has exclusive access; add this lock to the read locks
      else
      {
        locks.addElement(lock);
      }
    }

    // Lock the children, if any
    int success = lockDBChildren(lock, exclusive, waitFlag, waitDuration);

    // If the children could not be locked, then unlock this object
    if (success != LOCK)
    {
      if (exclusive)
      {
        this.exclusive = null;
        this.lockChildren = false;
      }
      else
        locks.removeElement(lock);

      // Unlock any of our children
      unlockDBChildren(lock);
      return NOLOCK;
    }

    // The desired lock was successful :-)!!
    return LOCK;
  }
}

class ttDate extends DBObject
{
  // Values for this object
  public int year;
  public int month;
  public int weekDay;
  public int day;
  public int hour;
  public int minute;
  public boolean includeWeekDay;
  public boolean includeTime;

  public ttDate()
  {
    super();
  }

  public int dataType()
  {
    return DATE;
  }

  public int add(ttStr name, int type, DataInputStream in, DataOutputStream out, boolean addChildren)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public int rename(ttStr oldName, ttStr newName)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public int delete(ttStr name)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public DBObject getChild(StringTokenizer location)
  {
    // We don't have any children
    return null;
  }

  public int lockDBChildren(Object lock, boolean exclusive, boolean waitFlag, int waitDuration)
  {
    // We don't have any children
    return LOCK;
  }

  public int unlockDBChildren(Object lock)
  {
    // We don't have any children
    return NOLOCK;
  }

  public void load(DataInputStream in, boolean includeChildren)
  {
    try {
      // Load our variables from the input stream
      year = in.readInt();
      month = in.readInt();
      day = in.readInt();
      weekDay = in.readInt();
      hour = in.readInt();
      minute = in.readInt();
      includeWeekDay = in.readBoolean();
      includeTime = in.readBoolean();
    }
    catch (IOException str) {}
  }

  public void save(DataOutputStream out, boolean includeChildren)
  {
    try {
      // Save our variables into the output stream
      out.writeInt(year);
      out.writeInt(month);
      out.writeInt(day);
      out.writeInt(weekDay);
      out.writeInt(hour);
      out.writeInt(minute);
      out.writeBoolean(includeWeekDay);
      out.writeBoolean(includeTime);
    }
    catch (IOException str) {}
  }
}

class ttDist extends DBObject
{
  // Values for this object
  private long english;  // in 1/16 inches
  private long metric;   // in millimeters
  public boolean useMetric;
  public int trial;
  public int points;
  public int lane;

  // Display flags
  public boolean includeHuge;   // kilometers and miles
  public boolean includeLarge;  // meters and yards
  public boolean includeMedium; // decimeters and feet
  public boolean includeSmall;  // centimeters and inches
  public boolean includeTiny;   // millimeters and fractions of inches
  public boolean asFraction;    // whether or not to display as a fraction

  public ttDist()
  {
    super();
  }

  public int dataType()
  {
    return DIST;
  }

  public int add(ttStr name, int type, DataInputStream in, DataOutputStream out, boolean addChildren)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public int rename(ttStr oldName, ttStr newName)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public int delete(ttStr name)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public DBObject getChild(StringTokenizer location)
  {
    // We don't have any children
    return null;
  }

  public int lockDBChildren(Object lock, boolean exclusive, boolean waitFlag, int waitDuration)
  {
    // We don't have any children
    return LOCK;
  }

  public int unlockDBChildren(Object lock)
  {
    // We don't have any children
    return NOLOCK;
  }

  public void load(DataInputStream in, boolean includeChildren)
  {
    try {
      metric = in.readLong();
      english = in.readLong();
      useMetric = in.readBoolean();
      trial = in.readInt();
      points = in.readInt();
      lane = in.readInt();
      includeHuge = in.readBoolean();
      includeLarge = in.readBoolean();
      includeMedium = in.readBoolean();
      includeSmall = in.readBoolean();
      includeTiny = in.readBoolean();
      asFraction = in.readBoolean();
    }
    catch (IOException str) {}
  }

  public void save(DataOutputStream out, boolean includeChildren)
  {
    try {
      out.writeLong(metric);
      out.writeLong(english);
      out.writeBoolean(useMetric);
      out.writeInt(trial);
      out.writeInt(points);
      out.writeInt(lane);
      out.writeBoolean(includeHuge);
      out.writeBoolean(includeLarge);
      out.writeBoolean(includeMedium);
      out.writeBoolean(includeSmall);
      out.writeBoolean(includeTiny);
      out.writeBoolean(asFraction);
    }
    catch (IOException str) {}
  }
}

class ttStr extends DBObject
{
  // Value for this object
  public String value;

  public ttStr()
  {
    super();
  }

  public int dataType()
  {
    return STR;
  }

  public int add(ttStr name, int type, DataInputStream in, DataOutputStream out, boolean addChildren)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public int rename(ttStr oldName, ttStr newName)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public int delete(ttStr name)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public DBObject getChild(StringTokenizer location)
  {
    // We don't have any children
    return null;
  }

  public int lockDBChildren(Object lock, boolean exclusive, boolean waitFlag, int waitDuration)
  {
    // We don't have any children
    return LOCK;
  }

  public int unlockDBChildren(Object lock)
  {
    // We don't have any children
    return NOLOCK;
  }

  public void load(DataInputStream in, boolean includeChildren)
  {
    try {
      // Load the string from the input
      value = in.readUTF();
/*      char str[];
      int len;
      in.readInt(len);len));
      in.read(str, 0, len);
      str[len] = '\0';
      value = new String(str);*/
    }
    catch (IOException str) {}
  }

  public void save(DataOutputStream out, boolean includeChildren)
  {
    try {
      // Save this string
      out.writeUTF(value);
/*      char str[] = value.toCharArray();
      int len = value.length();
      out.write(len);len));
      out.write(str, 0, len);*/
    }
    catch (IOException str) {}
  }
}

class ttTime extends DBObject
{
  // Values for this object
  public long value;
  public int trial;
  public int points;
  public int lane;

  // Display variables
  public boolean includeHours;
  public boolean includeMinutes;
  public boolean includeSeconds;
  public boolean includeCentiseconds;

  public ttTime()
  {
    super();
  }

  public int dataType()
  {
    return TIME;
  }

  public int add(ttStr name, int type, DataInputStream in, DataOutputStream out, boolean addChildren)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public int rename(ttStr oldName, ttStr newName)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public int delete(ttStr name)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public DBObject getChild(StringTokenizer location)
  {
    // We don't have any children
    return null;
  }

  public int lockDBChildren(Object lock, boolean exclusive, boolean waitFlag, int waitDuration)
  {
    // We don't have any children
    return LOCK;
  }

  public int unlockDBChildren(Object lock)
  {
    // We don't have any children
    return NOLOCK;
  }

  public void load(DataInputStream in, boolean includeChildren)
  {
    try {
      value = in.readLong();
      trial = in.readInt();
      points = in.readInt();
      lane = in.readInt();
      includeHours = in.readBoolean();
      includeMinutes = in.readBoolean();
      includeSeconds = in.readBoolean();
      includeCentiseconds = in.readBoolean();
    }
    catch (IOException str) {}
  }

  public void save(DataOutputStream out, boolean includeChildren)
  {
    try {
      out.writeLong(value);
      out.writeInt(trial);
      out.writeInt(points);
      out.writeInt(lane);
      out.writeBoolean(includeHours);
      out.writeBoolean(includeMinutes);
      out.writeBoolean(includeSeconds);
      out.writeBoolean(includeCentiseconds);
    }
    catch (IOException str) {}
  }
}

class ttEvent extends DBObject
{
  public final static int FIELDEVENT = 1;
  public final static int TRACKEVENT = 2;
  public final static int HEATSHEET = 3;
  public final static int RESERVED1 = 4;
  public final static int RESERVED2 = 5;

  public int scoreType;
  public Hashtable scores;
  public ttStr description;
  public ttDate day;
  public DBObject qualify;
  public int scoreIncrement;
  public int scoreHighest;
  public int scoreLowest;
  public int type;
  public boolean inTeamTotal;
  public boolean done;

  public ttEvent()
  {
    super();
    scores = new Hashtable();
  }

  public int dataType()
  {
    return EVENT;
  }

  public int add(ttStr name, int type, DataInputStream in, DataOutputStream out, boolean addChildren)
  {
    try
	{
	  // Add a new child
      DBObject obj = (DBObject) scores.get(name);
	  if (obj == null)
	  {
	    // Ignore the specified type; create what we are set to
	    switch (scoreType)
		{
	      case DIST :
		    obj = new ttDist();
			break;

		  case TIME :
		    obj = new ttTime();
			break;
		
		  default :
		    return ERROR;
		}

  	    // Notify the client of the object creation
		out.writeInt(OK);

		// Load the object and add it to the list
	    obj.load(in, addChildren);
		in.readInt();
	    scores.put(name, obj);
	    return OK;
	  }
	  
	  // Notify the client of the existant state of the object
	  out.writeInt(EXIST);
	  return EXIST;

	} catch (IOException str)
	{
	  return ERROR;
	}
  }

  public int rename(ttStr oldName, ttStr newName)
  {
	// Rename a child (they went to the courts and...)
    DBObject obj = (DBObject) scores.get(oldName);
	if (obj != null)
	{
	  scores.remove(oldName);
	  scores.put(newName, obj);
	  return OK;
    }
	return NOTEXIST;
  }

  public int delete(ttStr name)
  {
	// Delete the name from our list
    DBObject obj = (DBObject) scores.get(name);
	if (obj != null)
	{
	  scores.remove(name);
	  return OK;
    }
	return NOTEXIST;
  }

  public DBObject getChild(StringTokenizer location)
  {
    // Find any children with the given name (ours are Team|Participant)
    String str = location.nextToken();
    if (location.hasMoreTokens())
    {
      str = str + "|" + location.nextToken();
      ttStr cmp = new ttStr();
      cmp.value = str;
      return (DBObject) scores.get(cmp);
    }
    return null;
  }

  public int lockDBChildren(Object lock, boolean exclusive, boolean waitFlag, int waitDuration)
  {
    // Lock our children
    Enumeration list;
    int access = LOCK;
    list = scores.keys();
    while(list.hasMoreElements() && (access == LOCK))
    {
      // Get this participant's name
      ttStr str = (ttStr) list.nextElement();

      // Get the object and lock it
      DBObject obj = (DBObject) scores.get(str);
      access = obj.lockObject(lock, exclusive, true, waitFlag, waitDuration);
    }

    return access;
  }

  public int unlockDBChildren(Object lock)
  {
    // Unlock our children
    Enumeration list;
    list = scores.keys();
    while(list.hasMoreElements())
    {
      // Get this participant's name
      ttStr str = (ttStr) list.nextElement();

      // Get the object and unlock it
      DBObject obj = (DBObject) scores.get(str);
      obj.unlockObject(lock, true);
    }

    return NOLOCK;
  }

  public void load(DataInputStream in, boolean includeChildren)
  {
    try
    {
      // Clear the hashtable
      scores.clear();

      // Load the scores
      int i, numScores;
      numScores = in.readInt();
      scoreType = in.readInt();
      for (i = 1; i <= numScores; i++)
      {
        // Get this participant's name
        ttStr str = new ttStr();
        DBObject obj;
        str.load(in, includeChildren);

        if (includeChildren)
        {
          // Create the object
          switch(scoreType)
          {
            case DIST : obj = new ttDist(); break;
            case TIME : obj = new ttTime(); break;
          }

          // Load the object
          obj.load(in, includeChildren);

          // Add this score to the hashtable
          scores.put(str, obj);
        }
      }

      // Load the description and day
      description = new ttStr();
      description.load(in, includeChildren);
      day = new ttDate();
      day.load(in, includeChildren);

      // Load the qualifying object (for heat sheets)
      switch(scoreType)
      {
        case DIST : qualify = new ttDist(); break;
        case TIME : qualify = new ttTime(); break;
      }
      qualify.load(in, includeChildren);

      // Load the event's information
      scoreIncrement = in.readInt();
      scoreHighest = in.readInt();
      scoreLowest = in.readInt();
      type = in.readInt();
      inTeamTotal = in.readBoolean();
      done = in.readBoolean();
    } catch (IOException str) {  return; }
  }

  public void save(DataOutputStream out, boolean includeChildren)
  {
    try
    {
      // Save the score information
      int numScores = scores.size();
      out.writeInt(numScores);
      out.writeInt(scoreType);

      // Save the scores
      Enumeration list;
      for (list = scores.keys(); list.hasMoreElements();)
      {
        // Get this participant's name
        ttStr str = (ttStr) list.nextElement();
        str.save(out, includeChildren);

        if (includeChildren)
        {
          // Get the object and save it
          DBObject obj = (DBObject) scores.get(str);
          obj.save(out, includeChildren);
        }
      }

      // Save the description and day
      description.save(out, includeChildren);
      day.save(out, includeChildren);

      // Save the qualifying object (for heat sheets)
      qualify.save(out, includeChildren);

      // Save the event's information
      out.writeInt(scoreIncrement);
      out.writeInt(scoreHighest);
      out.writeInt(scoreLowest);
      out.writeInt(type);
      out.writeBoolean(inTeamTotal);
      out.writeBoolean(done);
    } catch (IOException str) {  return; }
  }
}

class Team extends DBObject
{
  public Vector events;
  public Hashtable participants;
  public ttStr description;
  public int teamTotal;

  public Team()
  {
    super();
    events = new Vector();
    participants = new Hashtable();
  }

  public int dataType()
  {
    return TEAM;
  }

  public int add(ttStr name, int type, DataInputStream in, DataOutputStream out, boolean addChildren)
  {
    try
	{
	  // Add a new child
      DBObject obj = (DBObject) participants.get(name);
	  if (obj == null)
	  {
	    // Ignore the specified type; always create a participant
  	    obj = new Participant();

  	    // Notify the client of the object creation
		out.writeInt(OK);

		// Load the object and add it to the list
	    obj.load(in, addChildren);
		in.readInt();
	    participants.put(name, obj);
	    return OK;
	  }
	  
	  // Notify the client of the existant state of the object
	  out.writeInt(EXIST);
	  return EXIST;

	} catch (IOException str)
	{
	  return ERROR;
	}
  }

  public int rename(ttStr oldName, ttStr newName)
  {
	// Rename a child (they went to the courts and...)
    DBObject obj = (DBObject) participants.get(oldName);
	if (obj != null)
	{
	  participants.remove(oldName);
	  participants.put(newName, obj);
	  return OK;
    }
	return NOTEXIST;
  }

  public int delete(ttStr name)
  {
	// Delete the name from our list
    DBObject obj = (DBObject) participants.get(name);
	if (obj != null)
	{
	  participants.remove(name);
	  return OK;
    }
	return NOTEXIST;
  }

  public DBObject getChild(StringTokenizer location)
  {
    // Find any children with the given name
    ttStr str = new ttStr();
    str.value = location.nextToken();
    return (DBObject) participants.get(str);
  }

  public int lockDBChildren(Object lock, boolean exclusive, boolean waitFlag, int waitDuration)
  {
    // Lock our children
    Enumeration list;
    int access = LOCK;
    list = participants.keys();
    while(list.hasMoreElements() && (access == LOCK))
    {
      // Get this participant's name
      ttStr str = (ttStr) list.nextElement();

      // Get the object and lock it
      DBObject obj = (DBObject) participants.get(str);
      access = obj.lockObject(lock, exclusive, true, waitFlag, waitDuration);
    }

    return access;
  }

  public int unlockDBChildren(Object lock)
  {
    // Unlock our children
    Enumeration list;
    list = participants.keys();
    while(list.hasMoreElements())
    {
      // Get this participant's name
      ttStr str = (ttStr) list.nextElement();

      // Get the object and unlock it
      DBObject obj = (DBObject) participants.get(str);
      obj.unlockObject(lock, true);
    }

    return NOLOCK;
  }

  public void load(DataInputStream in, boolean includeChildren)
  {
    try
    {
      // Load the event pointers
      int i, numEvents;
      events.removeAllElements();
      numEvents = in.readInt();
      for (i = 1; i <= numEvents; i++)
      {
        // Get this event's value
        ttStr str = new ttStr();
        str.load(in, includeChildren);

        // Add this score to the vector
        events.addElement(str);
      }

      // Load the participants for this team
      int numParticipants;
      participants.clear();
      numParticipants = in.readInt();
      for (i = 1; i <= numParticipants; i++)
      {
        // Get this participant's name
        ttStr str = new ttStr();
        DBObject obj;
        str.load(in, includeChildren);

        if (includeChildren)
        {
          // Load the object
          obj = new Participant();
          obj.load(in, includeChildren);

          // Add this score to the hashtable
          participants.put(str, obj);
        }
      }

      // Load the description
      description = new ttStr();
      description.load(in, includeChildren);

      // Load the team's information
      teamTotal = in.readInt();
    } catch (IOException str) {  return; }
  }

  public void save(DataOutputStream out, boolean includeChildren)
  {
    try
    {
      // Save the event pointers
      int numEvents = events.size();
      out.writeInt(numEvents);
      Enumeration list;
      for (list = events.elements(); list.hasMoreElements();)
      {
        // Get this event's name
        ttStr str = (ttStr) list.nextElement();
        str.save(out, includeChildren);
      }

      // Save the participants
      int numParticipants = participants.size();
      out.writeInt(numParticipants);
      for (list = participants.keys(); list.hasMoreElements();)
      {
        // Get this participant's name
        ttStr str = (ttStr) list.nextElement();
        str.save(out, includeChildren);

        if (includeChildren)
        {
          // Get the object and save it
          DBObject obj = (DBObject) participants.get(str);
          obj.save(out, includeChildren);
        }
      }

      // Save the description
      description.save(out, includeChildren);

      // Save the event's information
      out.writeInt(teamTotal);
    } catch (IOException str) {  return; }
  }
}

class Participant extends DBObject
{
  public Vector events;
  public ttStr description;
  public int points;

  public Participant()
  {
    super();
    events = new Vector();
  }

  public int dataType()
  {
    return PARTICIPANT;
  }

  public int add(ttStr name, int type, DataInputStream in, DataOutputStream out, boolean addChildren)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public int rename(ttStr oldName, ttStr newName)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public int delete(ttStr name)
  {
	// We don't have any children; return NOTEXIST
	return NOTEXIST;
  }

  public DBObject getChild(StringTokenizer location)
  {
    // We don't have any children
    return null;
  }

  public int lockDBChildren(Object lock, boolean exclusive, boolean waitFlag, int waitDuration)
  {
    // We don't have any children
    return LOCK;
  }

  public int unlockDBChildren(Object lock)
  {
    // We don't have any children
    return NOLOCK;
  }

  public void load(DataInputStream in, boolean includeChildren)
  {
    try
    {
      // Load the event pointers
      int i, numEvents;
      numEvents = in.readInt();
      for (i = 1; i <= numEvents; i++)
      {
        // Get this event's value
        ttStr str = new ttStr();
        str.load(in, includeChildren);

        // Add this score to the vector
        events.addElement(str);
      }

      // Load the description
      description = new ttStr();
      description.load(in, includeChildren);

      // Load the participant's information
      points = in.readInt();
    } catch (IOException str) {  return; }
  }

  public void save(DataOutputStream out, boolean includeChildren)
  {
    try
    {
      // Save the event pointers
      int numEvents = events.size();
      out.writeInt(numEvents);
      Enumeration list;
      for (list = events.elements(); list.hasMoreElements();)
      {
        // Get this event's name
        ttStr str = (ttStr) list.nextElement();
        str.save(out, includeChildren);
      }

      // Save the description
      description.save(out, includeChildren);

      // Save the participant's information
      out.writeInt(points);
    } catch (IOException str) {  return; }
  }
}

class TrackMeet extends DBObject
{
  public Hashtable events;
  public ttStr description;

  public TrackMeet()
  {
    super();
    events = new Hashtable();
  }

  public int dataType()
  {
    return TRACKMEET;
  }

  public int add(ttStr name, int type, DataInputStream in, DataOutputStream out, boolean addChildren)
  {
    try
	{
	  // Add a new child
      DBObject obj = (DBObject) events.get(name);
	  if (obj == null)
	  {
	    // Ignore the specified type; always create a participant
  	    obj = new ttEvent();

  	    // Notify the client of the object creation
		out.writeInt(OK);

		// Load the object and add it to the list
	    obj.load(in, addChildren);
		in.readInt();
	    events.put(name, obj);
	    return OK;
	  }
	  
	  // Notify the client of the existant state of the object
	  out.writeInt(EXIST);
	  return EXIST;

	} catch (IOException str)
	{
	  return ERROR;
	}
  }

  public int rename(ttStr oldName, ttStr newName)
  {
	// Rename a child (they went to the courts and...)
    DBObject obj = (DBObject) events.get(oldName);
	if (obj != null)
	{
	  events.remove(oldName);
	  events.put(newName, obj);
	  return OK;
    }
	return NOTEXIST;
  }

  public int delete(ttStr name)
  {
	// Delete the name from our list
    DBObject obj = (DBObject) events.get(name);
	if (obj != null)
	{
	  events.remove(name);
	  return OK;
    }
	return NOTEXIST;
  }

  public DBObject getChild(StringTokenizer location)
  {
    // Find any children with the given name
    ttStr str = new ttStr();
    str.value = location.nextToken();
    return (DBObject) events.get(str);
  }

  public int lockDBChildren(Object lock, boolean exclusive, boolean waitFlag, int waitDuration)
  {
    // Lock our children
    Enumeration list;
    int access = LOCK;
    list = events.keys();
    while(list.hasMoreElements() && (access == LOCK))
    {
      // Get this event's name
      ttStr str = (ttStr) list.nextElement();

      // Get the object and lock it
      DBObject obj = (DBObject) events.get(str);
      access = obj.lockObject(lock, exclusive, true, waitFlag, waitDuration);
    }

    return access;
  }

  public int unlockDBChildren(Object lock)
  {
    // Unlock our children
    Enumeration list;
    list = events.keys();
    while(list.hasMoreElements())
    {
      // Get this event's name
      ttStr str = (ttStr) list.nextElement();

      // Get the object and unlock it
      DBObject obj = (DBObject) events.get(str);
      obj.unlockObject(lock, true);
    }

    return NOLOCK;
  }

  public void load(DataInputStream in, boolean includeChildren)
  {
    try
    {
      // Clear the hashtable
      events.clear();

      // Load the events
      int i, numEvents;
      numEvents = in.readInt();
      for (i = 1; i <= numEvents; i++)
      {
        // Get this event's name
        ttStr str = new ttStr();
        DBObject obj;
        str.load(in, includeChildren);

        if (includeChildren)
        {
          // Load the object
          obj = new ttEvent();
          obj.load(in, includeChildren);

          // Add this score to the hashtable
          events.put(str, obj);
        }
      }

      // Load the description
      description = new ttStr();
      description.load(in, includeChildren);

    } catch (IOException str) {  return; }
  }

  public void save(DataOutputStream out, boolean includeChildren)
  {
    try
    {
      // Save event information
      int numEvents = events.size();
      out.writeInt(numEvents);

      // Save the scores
      Enumeration list;
      for (list = events.keys(); list.hasMoreElements();)
      {
        // Get this event's name
        ttStr str = (ttStr) list.nextElement();
        str.save(out, includeChildren);

        if (includeChildren)
        {
          // Get the object and save it
          DBObject obj = (DBObject) events.get(str);
          obj.save(out, includeChildren);
        }
      }

      // Save the description
      description.save(out, includeChildren);

    } catch (IOException str) {  return; }
  }
}

class ttRoot extends DBObject
{
  public File hFile;
  public Hashtable teams;
  public Hashtable trackMeets;
  public ttStr description;
  public int saveFrequency;

  public ttRoot()
  {
    super();
    teams = new Hashtable();
    trackMeets = new Hashtable();
    description = new ttStr();
    description.value = "TrackTrek '97 Beta1";
    saveFrequency = 0;
  }

  public int dataType()
  {
    return ROOT;
  }

  public int add(ttStr name, int type, DataInputStream in, DataOutputStream out, boolean addChildren)
  {
    try
	{
	  // Add a new child
      DBObject obj = (DBObject) teams.get(name);
	  if (obj == null)
		obj = (DBObject) trackMeets.get(name);

	  if (obj == null)
	  {
	    // Create the specified type of object
		switch (type)
		{
		  case TEAM :
			obj = new Team();
			break;

		  case TRACKMEET :
			obj = new TrackMeet();
			break;

		  default :
			return ERROR;
		}

  	    // Notify the client of the object creation
		out.writeInt(OK);

		// Load the object and its OK
	    obj.load(in, addChildren);
		in.readInt();

		// Add this object to the appropriate list
		switch (type)
		{
		  case TEAM :
			teams.put(name, obj);
			break;

		  case TRACKMEET :
			trackMeets.put(name, obj);
			break;
		}

		// Return OK
	    return OK;
	  }
	  
	  // Notify the client of the existant state of the object
	  out.writeInt(EXIST);
	  return EXIST;

	} catch (IOException str)
	{
	  return ERROR;
	}
  }

  public int rename(ttStr oldName, ttStr newName)
  {
	// Rename a child (they went to the courts and...)
	// Try the teams
    DBObject obj = (DBObject) teams.get(oldName);
	if (obj != null)
	{
	  teams.remove(oldName);
	  teams.put(newName, obj);
	  return OK;
    }

	// Try the track meets
	obj = (DBObject) trackMeets.get(oldName);
	if (obj != null)
	{
	  trackMeets.remove(oldName);
	  trackMeets.put(newName, obj);
	  return OK;
    }
	return NOTEXIST;
  }

  public int delete(ttStr name)
  {
	// Delete the name from our list
	// Try the teams
    DBObject obj = (DBObject) teams.get(name);
	if (obj != null)
	{
	  teams.remove(name);
	  return OK;
    }

    // Try the track meets
	obj = (DBObject) trackMeets.get(name);
	if (obj != null)
	{
	  trackMeets.remove(name);
	  return OK;
    }
	return NOTEXIST;
  }

  public DBObject getChild(StringTokenizer location)
  {
    // Find any children with the given name
    ttStr str = new ttStr();
    DBObject obj;
    str.value = location.nextToken();
    obj = (DBObject) teams.get(str);
    if (obj == null)
      obj = (DBObject) trackMeets.get(str);
    return obj;
  }

  public int lockDBChildren(Object lock, boolean exclusive, boolean waitFlag, int waitDuration)
  {
    // Lock our children
    Enumeration list;
    int access = LOCK;
    list = teams.keys();
    while(list.hasMoreElements() && (access == LOCK))
    {
      // Get this team's name
      ttStr str = (ttStr) list.nextElement();

      // Get the object and lock it
      DBObject obj = (DBObject) teams.get(str);
      access = obj.lockObject(lock, exclusive, true, waitFlag, waitDuration);
    }

    list = trackMeets.keys();
    while(list.hasMoreElements() && (access == LOCK))
    {
      // Get this track meet's name
      ttStr str = (ttStr) list.nextElement();

      // Get the object and lock it
      DBObject obj = (DBObject) trackMeets.get(str);
      access = obj.lockObject(lock, exclusive, true, waitFlag, waitDuration);
    }

    return access;
  }

  public int unlockDBChildren(Object lock)
  {
    // Unlock our children
    Enumeration list;
    list = teams.keys();
    while(list.hasMoreElements())
    {
      // Get this team's name
      ttStr str = (ttStr) list.nextElement();

      // Get the object and unlock it
      DBObject obj = (DBObject) teams.get(str);
      obj.unlockObject(lock, true);
    }

    list = trackMeets.keys();
    while(list.hasMoreElements())
    {
      // Get this track meet's name
      ttStr str = (ttStr) list.nextElement();

      // Get the object and unlock it
      DBObject obj = (DBObject) trackMeets.get(str);
      obj.unlockObject(lock, true);
    }

    return NOLOCK;
  }

  public void load(DataInputStream in, boolean includeChildren)
  {
    try
    {
      // Clear the hashtables
      teams.clear();
      trackMeets.clear();

      // Load the teams
      int i, numTeams;
      numTeams = in.readInt();
      for (i = 1; i <= numTeams; i++)
      {
        // Get this team's name
        ttStr str = new ttStr();
        DBObject obj;
        str.load(in, includeChildren);

        if (includeChildren)
        {
          // Load the object
          obj = new Team();
          obj.load(in, includeChildren);

          // Add this score to the hashtable
          teams.put(str, obj);
        }
      }

      // Load the track meets
      int numTrackMeets;
      numTrackMeets = in.readInt();
      for (i = 1; i <= numTrackMeets; i++)
      {
        // Get this track meet's name
        ttStr str = new ttStr();
        DBObject obj;
        str.load(in, includeChildren);

        if (includeChildren)
        {
          // Load the object
          obj = new TrackMeet();
          obj.load(in, includeChildren);

          // Add this score to the hashtable
          trackMeets.put(str, obj);
        }
      }

      // Load the description
      description = new ttStr();
      description.load(in, includeChildren);

      // Load the root variables
      saveFrequency = in.readInt();
    } catch (IOException str) {  return; }
  }

  public void save(DataOutputStream out, boolean includeChildren)
  {
    try
    {
      // Save team information
      int numTeams = teams.size();
      out.writeInt(numTeams);

      // Save the teams
      Enumeration list;
      for (list = teams.keys(); list.hasMoreElements();)
      {
        // Get this team's name
        ttStr str = (ttStr) list.nextElement();
        str.save(out, includeChildren);

        if (includeChildren)
        {
          // Get the object and save it
          DBObject obj = (DBObject) teams.get(str);
          obj.save(out, includeChildren);
        }
      }

      // Save track meet information
      int numTrackMeets = trackMeets.size();
      out.writeInt(numTrackMeets);

      // Save the track meets
      for (list = trackMeets.keys(); list.hasMoreElements();)
      {
        // Get this track meet's name
        ttStr str = (ttStr) list.nextElement();
        str.save(out, includeChildren);

        if (includeChildren)
        {
          // Get the object and save it
          DBObject obj = (DBObject) trackMeets.get(str);
          obj.save(out, includeChildren);
        }
      }

      // Save the description
      description.save(out, includeChildren);

      // Save the root variables
      out.writeInt(saveFrequency);
    } catch (IOException str) {  return; }
  }

  public boolean saveToFile()
  {
    try { try {
      // Initialize the streams
      FileOutputStream fileOut = new FileOutputStream(hFile);
      DataOutputStream out = new DataOutputStream(fileOut);

      // Write the header
      String header = "TrackTrek '97  version: 1.00.97a  program: TrackTrek Server (tt97a)";
      out.writeUTF(header);

      // Write an escape character
      out.write(27);

      // Save ourself to this file
      save(out, true);

      // Done saving :-)  -- Close the file
      fileOut.close();
    } catch (IOException str) { return false; } } catch (FileNotFoundException str) { return false; }

    // Everything was great!  Return true
    return true;
  }

  public boolean loadFromFile(String file)
  {
    // Open the file provided; open TrackTrek.db if none specified
    if (file == null)
      file = "TrackTrek.db";
    hFile = new File(file);

    try { try {
      // Initialize the streams
      FileInputStream fileIn = new FileInputStream(hFile);
      DataInputStream in = new DataInputStream(fileIn);

      // Check the header
      String header = in.readUTF();

      // If the header is not the first version type, return false
      if (header.startsWith("TrackTrek '97  version: 1.0"))
        return false;

      // Read the escape character
      in.read();

      // Load ourself from this file
      load(in, true);

      // Done reading :-)  -- Close the file, for now
      fileIn.close();
    } catch (IOException str) { return false; } } catch (FileNotFoundException str) { return false; }

    // Everything was great!  Return true
    return true;
  }
}


// Server Threads/Windows-----------------------------

class ClientThread extends Thread
{
  public Socket conn;
  public DataInputStream in;
  public DataOutputStream out;
  public ttRoot root;
  public Vector dbLocks;

  public final static int OK = 1;
  public final static int ERROR = 2;
  public final static int PING = 3;
  public final static int READ = 4;
  public final static int WRITE = 5;
  public final static int CLOSE = 6;
  public final static int LOCK = 7;
  public final static int NOLOCK = 9;
  public final static int NOTEXIST = 10;
  public final static int UNLOCK = 11;
  public final static int EXIST = 12;
  public final static int ADD = 14;
  public final static int RENAME = 15;
  public final static int DELETE = 16;

  // unimplemented (user access messages; right now, everyone has full access)
  public final static int DENIED = 20;
  public final static int AUTH = 25;

  public ClientThread(Socket conn, ttRoot root)
  {
    super();
    try
    {
      this.conn = conn;
      in = new DataInputStream(conn.getInputStream());
      out = new DataOutputStream(conn.getOutputStream());
      this.root = root;
      dbLocks = new Vector();
    } catch (IOException str) {}
  }

  public void run()
  {
    // Continually read from the port until the remote requests a close
    int func;
    try
    {
      while ((func = in.readInt()) != CLOSE)
      {
        switch(func)
        {
          case PING :
          {
            // Return the ping request with OK, PING, PING, PING, "tt97a", OK
            ttStr version = new ttStr();
            version.value = new String("TrackTrek '97  version: 1.00.97a  program: TrackTrek Server (tt97a)");

            // Ping!...
            out.writeInt(OK);
            out.writeInt(PING);
            out.writeInt(PING);
            out.writeInt(PING);
            version.save(out, false);
            out.writeInt(OK);
            break;
          }

		  case ADD :
          {
			// Add a child to a parent
            ttStr location = new ttStr();
            location.load(in, false);
            ttStr name = new ttStr();
            name.load(in, false);
			int type = in.readInt();
			boolean addChildren = in.readBoolean();
            DBObject obj = DBObject.objectFromPointer(root, location.value);

			// If the object exist, then add a child
			if (obj != null)
			{
			  obj.add(name, type, in, out, addChildren);
			}
			else
			  out.writeInt(NOTEXIST);
			break;
          }

		  case RENAME :
          {
			// Rename a child
            ttStr location = new ttStr();
            location.load(in, false);
            ttStr oldName = new ttStr();
            oldName.load(in, false);
            ttStr newName = new ttStr();
            newName.load(in, false);
            DBObject obj = DBObject.objectFromPointer(root, location.value);

			// If the object exist, then rename a child
			if (obj != null)
			{
			  out.writeInt(obj.rename(oldName, newName));
			}
			else
			  out.writeInt(NOTEXIST);
			break;
          }

		  case DELETE :
          {
			// Delete a child
            ttStr location = new ttStr();
            location.load(in, false);
            ttStr name = new ttStr();
            name.load(in, false);
            DBObject obj = DBObject.objectFromPointer(root, location.value);

			// If the object exist, then delete a child
			if (obj != null)
			{
			  out.writeInt(obj.delete(name));
			}
			else
			  out.writeInt(NOTEXIST);
			break;
          }

          case READ :
          {
            // Read a database member
            ttStr location = new ttStr();
            location.load(in, false);
            boolean readChildren = in.readBoolean();
            DBObject obj = DBObject.objectFromPointer(root, location.value);

            // Perform a read and return the result if the object exists
            if (obj != null)
            {
              // Return the object info and return OK
              out.writeInt(OK);
              obj.save(out, readChildren);
              out.writeInt(OK);
            }
            else
              out.writeInt(NOTEXIST);
            break;
          }
   
          case WRITE :
          {
            // Write/Create a database member
            ttStr location = new ttStr();
            location.load(in, false);
            boolean writeChildren = in.readBoolean();
            DBObject obj = DBObject.objectFromPointer(root, location.value);

            if (obj == null)
            {
              // If the object does not exist, return NOTEXIST
              out.writeInt(NOTEXIST);
            }
            else
            {
              // Check for exclusive access
              if (obj.exclusive == this)
              {
                // Check for child access
                if (writeChildren)
                {
                  if (!obj.lockChildren)
                  {
                    // Trying to write to a child without access
                    out.writeInt(NOLOCK);
                  }
                  else
                  {
                    // Return OK and write the data
                    out.writeInt(OK);
                    obj.load(in, true);
					in.readInt();
                  }
                }
                else
                {
                  // Return OK and write the data; then read the OK from client
                  out.writeInt(OK);
                  obj.load(in, false);
                  in.readInt();
                }
              }
              else
              {
                // No exclusive access; return NOLOCK
                out.writeInt(NOLOCK);
              }
            }
            break;
          }

		  case UNLOCK :
          {
			// Unlock a database leaf/branch
			ttStr location = new ttStr();
			location.load(in, false);
			boolean unlockChildren = in.readBoolean();
			DBObject obj = DBObject.objectFromPointer(root, location.value);
			if (obj == null)
			{
              // If the object does not exist, return NOTEXIST
              out.writeInt(NOTEXIST);
            }
            else
            {
			  // Unlock the children and return its success/failure
			  if (obj.unlockObject(this, unlockChildren) == NOLOCK)
				out.writeInt(NOLOCK);
			  else
			    out.writeInt(UNLOCK);
			}
			break;
          }

          case LOCK :
          {
            // Lock a database leaf/branch
            boolean waitFlag = in.readBoolean();
            int waitDuration = in.readInt();
            ttStr location = new ttStr();
            location.load(in, false);
            boolean lockChildren = in.readBoolean();
            boolean exclusive = in.readBoolean();
            DBObject obj = DBObject.objectFromPointer(root, location.value);

            if (obj == null)
            {
              // If the object does not exist, return NOTEXIST
              out.writeInt(NOTEXIST);
            }
            else
            {
              // Check the parents
              if (DBObject.checkParents(root, location.value) == LOCK)
              {
                // Lock the object
                int success = obj.lockObject(this, exclusive, lockChildren, waitFlag, waitDuration);

                // If locked, add this lock to our local lock table
                if (success == LOCK)
                  dbLocks.addElement(obj);

                // Return the success flag
                out.writeInt(success);
              }
              else
              {
                // Parents have lock over us; return NOLOCK
                out.writeInt(NOLOCK);
              }
            }
            break;
          }
        }
      }

      // Acknowledge the close
      if (func == CLOSE)
        out.writeInt(OK);

      // Close this connection
      conn.close();
    }
    catch (IOException str)
    {
      // We're no longer connected; close anyway
      try { conn.close(); } catch (IOException str2) {}
    }

    // Free any remaining locks
    Enumeration list = dbLocks.elements();
    DBObject obj;
    while (list.hasMoreElements())
    {
      // Unlock this object and any children it has from us
      obj = (DBObject) list.nextElement();
      obj.unlockObject(this, true);
    }

    // End this thread by simply returning from this function :-)
  }
}

class ExitButton extends Button
{
  public boolean pressed;

  public ExitButton()
  {
    super("Shut Down");
    pressed = false;
  }

  public boolean action(Event evt, Object what)
  {
    // If we were pressed, then set the pressed flag to true 
    if ((evt.id == Event.ACTION_EVENT) || (evt.id == Event.WINDOW_DESTROY))
      pressed = true;

    try { try {
      // Make a connection to ourselves so that we can exit
      Socket conn = new Socket("127.0.0.1", 3040);
      DataOutputStream out = new DataOutputStream(conn.getOutputStream());
      out.writeInt(ClientThread.CLOSE);
      conn.close();
    } catch (UnknownHostException str) {} } catch (IOException str) {}

    return true;
  }
}

public class TTServer implements Runnable
{
  public void run()
  {
  }

  public static void main(String args[])
  {
    // Create our window and add components to it
    Frame window = new Frame("TrackTrek '97 Server Beta1");;
	window.setBackground(new Color(192, 192, 192));
    GridBagLayout buttonLayout = new GridBagLayout();
	Panel panel = new Panel();
    Label status = new Label();
    Label statusLabel = new Label("Server Status:");
    BorderLayout layout = new BorderLayout();
    ExitButton userExit = new ExitButton();
    panel.setLayout(buttonLayout);
	panel.add(userExit);
	window.resize(320, 75);
    window.setLayout(layout);
    window.add("West", statusLabel);
    window.add("Center", status);
    window.add("East", panel);
    window.setResizable(false);
    window.show();
    status.setText("Starting server...");

    try
    {
      // Initialize Server
      ServerSocket server;
      server = new ServerSocket(3040, 200);

      // Load database into memory
      status.setText("Loading database...");
      ttRoot db = new ttRoot();
      db.loadFromFile(null);

      // Test locking
      status.setText("Testing locking mechanism...");
      db.lockObject(server, true, true, true, 150000);
      db.unlockObject(server, true);

      // Enter server loop
      status.setText("Serving");
      while (!userExit.pressed)
      {
        ClientThread newClient = new ClientThread(server.accept(), db);
        newClient.start();
      }

      // Exclusively lock the database
      status.setText("Obtaining exclusive access...");
      server.close();
      db.lockObject(db, true, true, true, 150000);

      // Save the database
      status.setText("Saving database...");
      db.saveToFile();
      status.setText("Shutting down...");
      db.unlockObject(db, true);
      System.exit(0);
    }
    catch (IOException str)
    {
      status.setText("Error loading!  Server not started.");

      // Wait until user exits
      while (!userExit.pressed)
        Thread.yield();
    }
  }
}
